package hu.bme.mit.ftsrg.healthcentre;

import hu.bme.mit.ftsrg.healthcentre.model.Doctor;
import hu.bme.mit.ftsrg.healthcentre.model.HealthCentreModel;
import hu.bme.mit.ftsrg.healthcentre.session.Session;

import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

public final class HealthCentreApplication {
	public static final String DATE_FORMAT_STRING = "yyyy-MM-dd HH:mm";
	public static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat(
			DATE_FORMAT_STRING);

	private final int port;
	private final File configXML;
	private final PrintStream out;
	private final PrintStream err;
	private HealthCentreModel model;

	public HealthCentreApplication(int port, File configXML, PrintStream out,
			PrintStream err) {
		this.port = port;
		this.configXML = configXML;
		this.out = out;
		this.err = err;
	}

	private void run() {
		out.println("===== HealthCentre Server =====");
		out.println("Run requested");
		out.println("  Desired port: " + port);
		out.println("  Configuration file: " + configXML);

		try {
			out.println("== Parsing configuration file ==");
			parseConfig();

			out.println("== Starting server ==");
			startServer();
		} catch (HealthCentreException e) {
			err.println("Exception: " + e.getMessage());

			if (e.getCause() != null) {
				e.getCause().printStackTrace(err);
			}
		}
	}

	private void parseConfig() throws HealthCentreException {
		try {
			model = new HealthCentreModel();

			// read and normalise XML
			DocumentBuilder builder = DocumentBuilderFactory.newInstance()
					.newDocumentBuilder();
			Document doc = builder.parse(configXML);
			doc.getDocumentElement().normalize();

			// validate root element
			Element rootElement = doc.getDocumentElement();
			if ("config".equals(rootElement.getNodeName())) {
				// get doctors
				NodeList doctorElements = rootElement
						.getElementsByTagName("doctor");

				if (doctorElements.getLength() == 0) {
					throw new HealthCentreException(
							"Configuration file: at least one doctor is required");
				}

				// get data for each doctor
				for (int i = 0; i < doctorElements.getLength(); i++) {
					Element doctorElement = (Element) doctorElements.item(i);

					Element nameElement = (Element) doctorElement
							.getElementsByTagName("name").item(0);
					Element sealNumberElement = (Element) doctorElement
							.getElementsByTagName("sealNumber").item(0);
					Element passwordElement = (Element) doctorElement
							.getElementsByTagName("password").item(0);
					Element availableDaysElement = (Element) doctorElement
							.getElementsByTagName("availableDays").item(0);
					Element availableTimesElement = (Element) doctorElement
							.getElementsByTagName("availableTimes").item(0);

					// no missing data
					if (nameElement == null || sealNumberElement == null
							|| availableDaysElement == null
							|| availableTimesElement == null) {
						throw new HealthCentreException(
								"Configuration file: the name, sealNumber, password "
										+ "availableDays and availableTimes elemnts "
										+ "must be defined for each doctor");
					}

					// get seal number
					int sealNumber = -1;
					try {
						sealNumber = Integer.parseInt(sealNumberElement
								.getTextContent().trim());
					} catch (NumberFormatException e) {
					}

					if (sealNumber <= 0) {
						throw new HealthCentreException(
								"Configuration file: the seal number must"
										+ " be positive number for each doctor");
					}

					// create Doctor object
					Doctor doctor = new Doctor(nameElement.getTextContent()
							.trim(), sealNumber, passwordElement
							.getTextContent().trim());

					// get days and times
					NodeList dayElements = availableDaysElement
							.getElementsByTagName("day");
					NodeList timeElements = availableTimesElement
							.getElementsByTagName("time");

					if (dayElements.getLength() == 0
							|| timeElements.getLength() == 0) {
						throw new HealthCentreException(
								"Configuration file: there must be at least 1 day "
										+ "and 1 time for each doctor");
					}

					String[] days = new String[dayElements.getLength()];
					String[] times = new String[timeElements.getLength()];

					for (int j = 0; j < dayElements.getLength(); j++) {
						days[j] = dayElements.item(j).getTextContent().trim();
					}

					for (int j = 0; j < timeElements.getLength(); j++) {
						times[j] = timeElements.item(j).getTextContent().trim();
					}

					// create available dates (~matrix)
					for (String day : days) {
						for (String time : times) {
							String dateString = day + " " + time;

							try {
								doctor.availableDates().add(
										DATE_FORMAT.parse(dateString));
							} catch (ParseException e) {
								throw new HealthCentreException(
										"Configuration file: invalid date string "
												+ dateString + " for doctor "
												+ doctor);
							}
						}
					}

					// save to the model
					model.doctors().put(doctor.getName(), doctor);
				}
			} else {
				throw new HealthCentreException(
						"Configuration file: the root element should be <config>");
			}
		} catch (ParserConfigurationException | SAXException | IOException e) {
			throw new HealthCentreException(
					"Unexpected exception occured when parsing"
							+ " the configuration file (see details)", e);
		}
	}

	private void startServer() throws HealthCentreException {
		// open socket
		try (ServerSocket serverSocket = new ServerSocket(port)) {
			out.println("Socket created");

			while (true) {
				// accept connections from the clients
				try {
					Socket socket = serverSocket.accept();
					out.println("Client connected: " + socket.getInetAddress()
							+ ":" + socket.getPort());

					// start session
					new Thread(new Session(socket, model, out, err)).start();
				} catch (IOException e) {
					throw new HealthCentreException(
							"Unexpected exception occured when creating"
									+ " a socket (see details)", e);
				}
			}
		} catch (IOException e) {
			throw new HealthCentreException(
					"Unexpected exception occured when starting"
							+ " the server socket (see details)", e);
		}
	}

	public static void main(String[] args) {
		// parse command line arguments
		int port = -1;
		File configXML = null;
		List<String> errors = new ArrayList<>();

		if (args.length == 2) {
			// port
			try {
				port = Integer.parseInt(args[0]);

				if (port <= 0 || port >= 65536) {
					errors.add("Invalid port: " + port);
				}
			} catch (NumberFormatException e) {
				errors.add("Not a number: " + port);
			}

			// configuration file
			configXML = new File(args[1]);

			if (!configXML.exists()) {
				errors.add("The configuration file does not exists");
			} else if (!configXML.canRead()) {
				errors.add("The configuration file cannot be read");
			}
		} else {
			errors.add("Missing parameters");
		}

		if (errors.size() > 0) {
			// print errors
			System.err.println("An error occured");

			for (String error : errors) {
				System.err.println("  " + error);
			}

			System.err
					.println("Usage: java -jar HealthCentre.jar <port> <configxml>");
		} else {
			// start application
			try {
				new HealthCentreApplication(port, configXML, System.out,
						System.err).run();
			} catch (Exception e) {
				System.err.println("Uncaught exception");
				System.err.println("  Class: " + e.getClass());
				System.err.println("  Message: " + e.getMessage());
				e.printStackTrace();
			}
		}
	}
}
